#!/usr/bin/env python

# Copyright and license: Please refer to LICENSE.TXT

import sys
import os
import os.path
import string
import re
import time
import signal
from xml.sax import saxutils
from threading import Lock

_selfpath=(os.path.dirname(os.path.realpath(__file__)).split('/'))[:-2]
_selfpath.append('pylibs')
_libspath='/'.join(_selfpath)
sys.path.append(_libspath)

from s2p import *
from s2p_usock import *

sys.path.append(CORE_PATH_SDK+"/python")

from s2p_base import *

class S2P_BRIDGE:
	
	def __init__(self):
		self._call_map_sc={}			# Map Skype/Chan
		self._call_map_cs={}			# Map Chan/Skype

		self._call_map_sc_garbage={}
		self._call_map_cs_garbage={}
		self._garbage_timeout=15

		self._options={}
		self._callbacks={}

		self.set_default_options()

		self._clean_lock=threading.Lock()
		
		# Handlers
		self._s2pc=S2P_CHANNELS()
		self._s2ps=S2P_SKYPE()

		# Internal callbacks
		self._s2ps.register_callback('message', self._on_skype_message_event)
		self._s2pc.register_callback('message', self._on_chan_message_event)

		self._s2ps.register_callback('call', self._on_skype_call_event)
		self._s2pc.register_callback('call', self._on_chan_call_event)
		
		self._s2ps.register_callback('vm', self._on_skype_vm_event)

		self._s2ps.register_callback('dtmf', self._on_skype_dtmf)
		self._s2pc.register_callback('dtmf', self._on_chan_dtmf)

	def _sc_on_garbage(self, callid):
		self._call_map_sc_garbage[callid]=time.time()+self._garbage_timeout

	def _cs_on_garbage(self, callid):
		self._call_map_cs_garbage[callid]=time.time()+self._garbage_timeout

	def _garbage_clean(self):
		#print "SC> "+str(len(self._call_map_sc.keys()))

		self._clean_lock.acquire()

		to_delete=[]
		for callid in self._call_map_sc_garbage.keys():
			if self._call_map_sc_garbage[callid] >= time.time(): continue
			if self._call_map_sc.has_key(callid):
				#self._call_map_sc[callid].send_finished()
				del self._call_map_sc[callid]

			to_delete.append(callid)
			
		for callid in to_delete: del self._call_map_sc_garbage[callid]

		#print "CS> "+str(len(self._call_map_cs.keys()))
		to_delete=[]
		for callid in self._call_map_cs_garbage.keys():
			if self._call_map_cs_garbage[callid] >= time.time(): continue
			if self._call_map_cs.has_key(callid):
				#self._call_map_cs[callid].send_finished()
				del self._call_map_cs[callid]

			to_delete.append(callid)
		
		for callid in to_delete:
			del self._call_map_cs_garbage[callid]

		self._clean_lock.release()

	def get_bridged_calls_sk(self):
		return self._call_map_sc
	
	def get_bridged_calls_ch(self):
		return self._call_map_cs

	def get_ch_call(self, sk_call_id):
		if not self._call_map_sc.has_key(sk_call_id): return 0
		return self._call_map_sc[sk_call_id]

	def get_sk_call(self, ch_call_id):
		if not self._call_map_cs.has_key(ch_call_id): return 0
		return self._call_map_cs[ch_call_id]

	# Link a skype call with a chan call
	# Manager parameter specifes a channel manager object
	def link(self, ch_call, sk_call, ch_manager=None, sk_manager=None):
		self._garbage_clean()
		#print sk_call.get_id()+" <-> "+ch_call.get_id()+": "+str(manager)
		
		# Duplicate call check
		if ch_call and sk_call:
			if self._call_map_cs.has_key(ch_call.get_id()) and self._call_map_sc.has_key(sk_call.get_id()): return
			
		else:
			if ch_call and self._call_map_cs.has_key(ch_call.get_id()): return
			if sk_call and self._call_map_sc.has_key(sk_call.get_id()): return
		
		self._call_map_sc[sk_call.get_id()]=ch_call
		self._call_map_cs[ch_call.get_id()]=sk_call

  		sk_call.set_tcp_capture(ch_call.get_tcp_playback())
		sk_call.set_tcp_playback(ch_call.get_tcp_capture())

		# Playback eventi
		if sk_manager:
			events=sk_manager.get_events_history(sk_call.get_id())
			for event in events: self._on_skype_call_event(sk_call, event['event-name'])

		if ch_manager:
			events=ch_manager.get_events_history(ch_call.get_id())
			for event in events: self._on_chan_call_event(ch_call, event['args']["event"])

	# Reserved internal callback function
	def _on_skype_vm_event(self, call, event):

		# Locate the linked call
		if not self._call_map_sc.has_key(call.get_id()): return
		ch_call=self._call_map_sc[call.get_id()]
		
		ch_call.send_finished()
		call.send_finished()
		return
		
		#if event == 'vm-blank': ch_call.send_audio_restart()
			
		call.set_tcp_capture(ch_call.get_tcp_playback())
		call.set_tcp_playback(ch_call.get_tcp_capture())
		
		if event == "vm-played":
			ch_call.send_recv_tone({'freq': '2000', 'seconds': '0.2'})
			
		if \
			event == "vm-played" or \
			event == "vm-recording" \
		:
			ch_call.send_mute_recv({'status': 'true'})

	# Reserved internal callback function
	def _on_skype_call_event(self, call, event):

		link_destroy=False

		# New call
		if event == "call-incoming-ringing":
			if self._callbacks.has_key('from-skype'):
				res=self._callbacks['from-skype'](call)
				if res[0]: self.link(res[0], call)

		# Locate the linked call
		if not self._call_map_sc.has_key(call.get_id()): return

		ch_call=self._call_map_sc[call.get_id()]

		# Failure events
		if event == "call-outgoing-failed":
			link_destroy=True
			
			if call.get('failure-reason') == "1":
				
				if call.get('pstn-status'):

					if call.get('pstn-status')[0:5] == "10404":
						ch_call.send_failed({'reason': 'not-exists'})
						
					elif call.get('pstn-status')[0:5] == "10502":
						ch_call.send_failed({'reason': 'bad-gateway'})
						
					elif call.get('pstn-status')[0:4] == "9402":
						ch_call.send_failed({'reason': 'credit'})
						
					elif call.get('pstn-status')[0:4] == "9403":
						ch_call.send_failed({'reason': 'forbidden'})
						
					elif call.get('pstn-status')[0:4] == "9407":
						ch_call.send_failed({'reason': 'forbidden'})
						
					elif call.get('pstn-status')[0:4] == "9408":
						ch_call.send_failed({'reason': 'bad-gateway'})
	
					else:
						ch_call.send_failed({'reason': 'bad-gateway'})
				
				else:
					ch_call.send_failed({'reason': 'bad-gateway'})

			elif call.get('failure-reason') == "2":		ch_call.send_failed({'reason': 'not-exists'})
			elif call.get('failure-reason') == "3":		ch_call.send_failed({'reason': 'unavailable'})
			elif call.get('failure-reason') == "4":		ch_call.send_failed({'reason': 'bad-gateway'})
			elif call.get('failure-reason') == "9":		ch_call.send_failed({'reason': 'service'})
			elif call.get('failure-reason') == "10":	ch_call.send_failed({'reason': 'service'})
			elif call.get('failure-reason') == "11":	ch_call.send_failed({'reason': 'service'})
			else:										ch_call.send_failed({'reason': 'service'})

		# Finish events
		if event == "call-outgoing-busy":
			link_destroy=True
			ch_call.send_failed({'reason': 'busy'})

		# Refused events
		if event == "call-outgoing-refused":
			link_destroy=True
			ch_call.send_failed({'reason': 'decline'})

		# Standard call events
		if event == "call-incoming-terminated" or event == "call-outgoing-terminated" or event == "call-crashed":
			link_destroy=True
			ch_call.send_finished()

		if event == "call-outgoing-answered":
			#ch_call.send_ringing()
			#ch_call.send_progress()
			ch_call.send_answer()

		if event == "call-outgoing-routing":
			if self._options["ring-on-routing"]: ch_call.send_ringing()
		
		if event == "call-outgoing-ringing":
			ch_call.send_ringing()

		# Early media
		if event == "call-outgoing-earlymedia":
			#ch_call.send_ringing()
			ch_call.send_progress()
			#ch_call.send_answer()

		# Problemi audio
		if event == "call-input-finished":
			if \
				call.get_status() != "finished" and \
				call.get_status() != "failed" and \
				call.get_status() != "cancelled" \
			:
				call.set_tcp_capture(ch_call.get_tcp_playback())
				call.set_tcp_playback(ch_call.get_tcp_capture())

		# On-Hold / Resume events
		if self._options['forward-onhold']: 
			if event == "call-incoming-remotehold" or event == "call-outgoing-remotehold":
				ch_call.send_hold()

			if event == "call-incoming-resumed" or event == "call-outgoing-resumed":
				ch_call.send_resume()

		if link_destroy: self._sc_on_garbage(call.get_id())

	# Reserved internal callback function
	def _on_chan_call_event(self, call, event):

		link_destroy=False

		# New call
		if event == "call-incoming-start":
			if self._callbacks.has_key('from-chan'):
				res=self._callbacks['from-chan'](call)
				if res and res[0]: self.link(call, res[0])

		# Locate the linked call
		if not self._call_map_cs.has_key(call.get_id()): return
		sk_call=self._call_map_cs[call.get_id()]

		# Standard call events
		if event == "call-incoming-terminated" or event == "call-outgoing-terminated":
			link_destroy=True
			sk_call.send_finished()

		if event == "call-outgoing-answered":
			sk_call.send_answer()

		# On-Hold / Resume events
		if self._options['forward-onhold']: 
			if event == "call-incoming-remotehold" or event == "call-outgoing-remotehold":
				sk_call.send_hold()

			if event == "call-incoming-resumed" or event == "call-outgoing-resumed":
				sk_call.send_resume()

		if link_destroy: self._cs_on_garbage(call.get_id())

	# Reserved internal callback function
	def _on_skype_dtmf(self, call, code):
		# Locate the linked call
		if not self._call_map_sc.has_key(call.get_id()): return
		ch_call=self._call_map_sc[call.get_id()]
		
		if self._options['forward-dtmf']:
			#ch_call.send_mute({"seconds": "2"})
			ch_call.send_dtmf({
				'call-id'	: call.get_id(),
				'code'		: code
			})

	# Reserved internal callback function
	def _on_chan_dtmf(self, call, code):
		# Locate the linked call
		if not self._call_map_cs.has_key(call.get_id()): return
		sk_call=self._call_map_cs[call.get_id()]

		if self._options['forward-dtmf']:
			sk_call.send_dtmf({
				'call-id'	: call.get_id(),
				'code'		: code
			})

	# Reserved internal callback function
	def _on_skype_message_event(self, message, event):
		#print message
		1

	# Reserved internal callback function
	def _on_chan_message_event(self, message, event):
		#print message
		1

	# Register callbacks
	def register_callback(self, name, callback_function):
		self._callbacks[name]=callback_function

	# Set the default configuration
	def set_default_options(self):
		self._options['forward-onhold']		= False
		self._options['forward-chat']		= True
		self._options['forward-dtmf']		= True
		self._options['ring-on-routing']	= False

	# Set one option
	def set_option(self, key, value):
		self._options[key]=value

	# Main thread function
	def start(self):
		self._s2ps.start()
		self._s2pc.start()

	# Stop the thread
	def stop(self):
		self._s2ps.stop()
		self._s2pc.stop()

# Used for threading
signal.signal(signal.SIGINT, signal.SIG_DFL)
